/*
 * The MethodIterator class iterates through a method or function passed by the programmer, over a
 * programmer-specified domain. This domain can be indicated via a list, a range (with an optional
 * step size), or a /pif_Iterator object.
 */

#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_ITERATOR)
pif_Iterator/MethodIterator
	var
		pif_Iterator/Domain
#else
Iterator/MethodIterator
	var
		Iterator/Domain
#endif

		ProcRef

		Object
		ProcName

		LibName
		FuncName

	proc
		_Call()
			if(ProcRef)
				return call(ProcRef)(arglist(args))
			else if(Object && ProcName)
				return call(Object, ProcName)(arglist(args))
			else if(LibName && FuncName)
				return call(LibName, FuncName)(arglist(args))

	/*
	 * Protocol implementation.
	 */

	New(/* ... */)
		// Several different argument formats.

		var
			// Possible position for domain arguments to start
			domain_start = 3

		// First, we check for the call-related arguments.

		if(istype(args[1], /datum) && istext(args[2]))
			// If using the ref format new /pif_Iterator/MethodIterator(Object, ProcName, ...
            // Domain)

			Object = args[1]
			ProcName = args[2]

		else if(ispath(args[1]) && istext(args[2]))
			// If using the ref format new /pif_Iterator/MethodIterator(LibName, FuncName, ...
            // Domain)

			LibName = args[1]
			FuncName = args[2]

		else
			// Using the ref format new /pif_Iterator/MethodIterator(ProcRef). We can not text for
            // this directly, using ispath as ispath(/prof/Foo) returns 0, but it's the only
            // possible option.

			ProcRef = args[1]

			// One less argument than the other call-related formats, so we decrease the domain
            // argument start position by 1.
			domain_start --

		// Next, we check for the domain-related arguments.

		if(args.len < domain_start)
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_ITERATOR)
			throw EXCEPTION("Invalid domain arguments type for /pif_Iterator/MethodIterator.")
#else
			throw EXCEPTION("Invalid domain arguments type for /Iterator/MethodIterator.")
#endif

#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_ITERATOR)
		else if(istype(args[domain_start + 0], /pif_Iterator))
#else
		else if(istype(args[domain_start + 0], /Iterator))
#endif
			// Using the domain format new /pif_Iterator/MethodIterator(..., new /pif_Iterator/Foo(...))
			Domain = args[domain_start + 0]

		else if(istype(args[domain_start + 0], /list))
			// Using the domain format new /pif_Iterator/MethodIterator(..., list(...))
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_ITERATOR)
			Domain = new /pif_Iterator/ForwardListIterator(args[domain_start + 0])
#else
			Domain = new /Iterator/ForwardListIterator(args[domain_start + 0])
#endif

		else if((args.len == (domain_start + 1)) && isnum(args[domain_start + 0]) && isnum(args[domain_start + 1]))
			// Usng the domain format new /pif_Iterator/MethodIterator(..., Start, End)
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_ITERATOR)
			Domain = new /pif_Iterator/RangeIterator(args[domain_start + 0], args[domain_start + 1])
#else
			Domain = new /Iterator/RangeIterator(args[domain_start + 0], args[domain_start + 1])
#endif

		else if((args.len == (domain_start + 2)) && isnum(args[domain_start + 0]) && isnum(args[domain_start + 1]) && isnum(args[domain_start + 2]))
			// Using the domain format new /pif_Iterator/MethodIterator(..., Start, End, StepSize)
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_ITERATOR)
			Domain = new /pif_Iterator/RangeIterator(args[domain_start + 0], args[domain_start + 1], args[domain_start + 2])
#else
			Domain = new /Iterator/RangeIterator(args[domain_start + 0], args[domain_start + 1], args[domain_start + 2])
#endif

		// If we've gotten this far, the entire argument format is simply unknown.

		else
#if !defined(PIF_NOPREFIX_GENERAL) && !defined(PIF_NOPREFIX_ITERATOR)
			throw EXCEPTION("Unknown argument format for /pif_Iterator/MethodIterator.")
#else
			throw EXCEPTION("Unknown argument format for /Iterator/MethodIterator.")
#endif

		return 1

	Initialize(/* ... */)
		return Domain.Initialize(arglist(args))

	Next(/* ... */)
		return Domain.Next(arglist(args))

	Step(step)
		return Domain.Step(step)

	Look()
		. = Domain.Look()
		if(. != null)
			return _Call(.)
		else
			return 0

	Peek(step = 1)
		. = Domain.Peek(step)
		if(. != null)
			return _Call(.)

	Done()
		return Domain.Done()

	Position()
		return Domain.Position()